/* ini.c
 *
 *  Scott Lawrence
 *  yorgle@gmail.com
 *
 *  This was brought forth from my "Turaco" project, circa 2003
 */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/******

 This module turns an INI file into a queryable model.  You
 can load files in to the internal structure then query them  This only
 currently supports reading in INI files, since that's all we need to
 do at this point.

 It will also return entries into a whitespace-seperated array as
 strings or integers.  ie:
	key = 1 2 3 4 5
	assert( key[2] = "3" )

 ******/

#include <stdio.h>
#include <stdlib.h> /* malloc, free, atoi */
#include <string.h> /* memset, strlen, strdup */
#ifdef SunOS
#include <strings.h> /* strlen, strdup */
#endif
#include "inistrings.h"   /* for my string manipulation functions */

#include "ini.h"

/*
 * ini_Destroy
 *
 * clears the internal structure
 */
void ini_Destroy( INIStruct * i )
{
    INIStruct * next;

    while( i )
    {
	next = i->next;

	if ( i->group )  free( i->group );
	if ( i->key )    free( i->key );
	if ( i->value )  free( i->value );

	free( i );
	i = next;
    }
}


/*
 * ini_Query
 * 
 *  finds data from the internal strucutre.
 *  the returned data is a char string allocated internally.
 *  do not free it externally!   
 */
char * ini_Query( INIStruct * i, char * group, char * key )
{
    INIStruct * curr = i;

    if( !group || !key || !i)
	return( NULL );

    while( curr )
    {
	if (    (!strcasecmp( curr->group, group ))
	     && (!strcasecmp( curr->key, key ))
	    )
	    return( curr->value );

	curr = curr->next;
    }
    return( NULL );
}

char * ini_Query_IntKey( INIStruct * i, char * group, int key )
{
	char buf[16];
	snprintf( buf, 16, "%d", key );
	return( ini_Query( i, group, buf ));
}


/*
 * ini_QueryInt
 *
 *  finds data from the internal strucutre.
 *  the returned data is an int
 */
int ini_QueryInt( INIStruct * i, char * group, char * key )
{
    char * r = ini_Query( i, group, key );
    if( r )
	return( atoi( r ));
    return( 0 );
}



/*
 * ini_QueryArrayString
 * 
 * finds the n'th element in the array found via an INI query
 *  do not free it externally!   
 */
char * ini_QueryArrayString( INIStruct * i, char * group, char * key, int index )
{
    char * r = ini_Query( i, group, key );
    if (r)
    {
	return( inistrings_whiteSpaceArray( r, index ) );
    }
    return( NULL );
}


/*
 * ini_QueryArrayInt
 * 
 * finds the n'th element in the array found via an INI query
 *  do not free it externally!   
 */
int ini_QueryArrayInt( INIStruct * i, char * group, char * key, int index )
{
    char * r = ini_Query( i, group, key );
    char * r2;
    if (r)
    {
	r2 = inistrings_whiteSpaceArray( r, index );
	if( r2 ) return( atoi( r2 ));
    }
    return( 0 );
}


/*
 * ini_GroupExists
 *
 * finds the passed in group, returns 1 if it was found, 0 if it was not
 */
int ini_GroupExists( INIStruct * i, char * group )
{
    INIStruct * curr = i;
    
    while( curr )
    {
	if( !strcmp( curr->group, group )) 
	{
	    return( 1 );
	}
	curr = curr->next;
    }
    return( 0 );
}


/*
 * ini_KeyExists
 *
 * finds the passed in group/key, returns 1 if it was found, 0 if it was not
 */
int ini_KeyExists( INIStruct * i, char * group, char * key )
{
    INIStruct * curr = i;
    
    while( curr )
    {
	if( !strcmp( curr->group, group )) 
	{
	    if( !strcmp( curr->key, key ))
	    {
		return( 1 );
	    }
	}
	curr = curr->next;
    }
    return( 0 );
}

int ini_IntKeyExists( INIStruct * i, char * group, int key )
{
	char buf[16];
	snprintf( buf, 16, "%d", key );
	return( ini_KeyExists( i, group, buf ));
}

/* 
 * ini_Dump
 *
 *  dump out the internal header list. (for debug use)
 */
void ini_Dump( INIStruct * i )
{
    INIStruct * curr = i;

    printf("Dump:\n");
    while( curr )
    {
	printf( "  [%s] %s \t->%s\n", curr->group, curr->key, curr->value );

	curr = curr->next;
    }
}



#define BUFSIZE (1024)

/*
 * ini_AddEntry
 *
 *  adds an entry into the ini list
 */
INIStruct * ini_AddEntry( INIStruct * i, char * group, char * key, char * value )
{
    INIStruct * n;
    INIStruct * t;

    n = (INIStruct *) malloc( sizeof( INIStruct ));
    if( n )
    {
	n->group = strdup( group );
	n->key   = strdup( key );
	n->value = strdup( value );
	n->next  = NULL;

	/* this is the easy way to add - prepend to the beginning */
	/*
		n->next = i;
		i = n;
	*/

	/* instead, lets' insert in order... */
	/* if there's no nodes, we're the first... */
	if( !i ) {
		return( n ); /* WE'RE NUMBER ONE! */
	}
	/* okay, find the end of the list. */
	t = i;
	while( t->next ) {
		t = t->next;
	}
	t->next = n; /* tack it onto the end */
    }
    return( i );
}


/*
** __fpeek
**
**  look at the next character without getting it.... sort of
*/
int __fpeek(FILE * fp)
{
    int c = getc(fp);
    ungetc(c, fp);
    return c;
}



/*
 * ini_ParseFile
 *
 *   will parse in the file "filename" into the internal structures
 */
INIStruct * ini_ParseFile( char * filename )
{
    int x;
    char buf[BUFSIZE];
    char * header = NULL;
    char * key    = NULL;
    char * value  = NULL;
    FILE * ifp    = NULL;
    INIStruct * i = NULL;

    if (!filename)
	return( NULL );

    ifp = fopen( filename, "rb" );
    if (!ifp)
	return( NULL );

    while (__fpeek(ifp) != EOF)
    {
	// clear the buffer
	memset( buf, '\0', BUFSIZE );
	// read in a line
	(void) fgets( buf, BUFSIZE, ifp );

	// check for a # sign in it, set it to a NULL
	(void) inistrings_makeFirstNull( buf, BUFSIZE, '#' );

	// zot leading and trailing space
	(void) inistrings_trimWhitespace( buf, BUFSIZE );

	// skip if there's nothing left
	if (0 == strlen( buf )) continue;

	// check for square brackets.  if so, 
	if ( buf[0] == '[' && buf[strlen(buf)-1] == ']' )
	{
	    //   set first and last character to ' '
	    buf[0] = buf[strlen(buf)-1] = ' ';

	    //   zot leading and trailing space
	    (void) inistrings_trimWhitespace( buf, BUFSIZE );

	    //   that is the header
	    if (header)  free( header );
	    header = strdup( buf );
	} else {
	    // find '='
	    x = inistrings_strchrIndex( buf, BUFSIZE, '=', 0 );

	    // if not found, go to the next line
	    if (x == -1) continue;

	    key = value = NULL;

	    //   char * value = buf[that address+1]
	    value = buf+x+1;

	    //   set buf[that address] = '\0'
	    buf[x] = '\0';
	    
	    //   char * key = buf[0]
	    key = buf;

	    //   zot leading and trailing space (buf) (value)
	    (void) inistrings_trimWhitespace( key, BUFSIZE );
	    (void) inistrings_trimWhitespace( value, BUFSIZE );

	    //   CB key-value 
	    i = ini_AddEntry( i, header, key, value );
	}

    }
    fclose(ifp);

    /* free any allocated memory */
    if (header)
	free( header );

    return( i );
}
